const std = @import("std");

const Command = @import("./Command.zig");

///
const Self = @This();

///
const CommandList = std.ArrayList(Command);

///
commands: CommandList,

/// next command to be applied by redoing
head: usize,

///
pub fn init(alloc: std.mem.Allocator) Self {
    return Self {
        .commands = CommandList.init(alloc),
        .head     = 0,
    };
}

///
pub fn deinit(self: *Self) void {
    for (self.commands.items) |*command| {
        command.deinit();
    }
    self.commands.deinit();
}

///
pub fn exec(self: *Self, command: Command) !void {
    if (self.head >= self.commands.items.len) {
        try self.commands.append(command);
    } else {
        try self.commands.resize(self.head + 1);
        self.commands.items[self.head] = command;
    }
    errdefer _ = self.commands.pop();

    self.head = self.commands.items.len;
    try command.apply();
}

///
pub fn undo(self: *Self) !void {
    if (self.head == 0) {
        return error.NoCommand;
    }
    self.head -= 1;
    errdefer self.head += 1;

    try self.commands.items[self.head].revert();
}

///
pub fn redo(self: *Self) !void {
    if (self.head >= self.commands.items.len) {
        return error.NoCommand;
    }
    try self.commands.items[self.head].apply();
    self.head += 1;
}

///
pub fn isUndoAvailable(self: *const Self) bool {
    return self.head > 0;
}
///
pub fn isRedoAvailable(self: *const Self) bool {
    return self.head < self.commands.items.len;
}

test {
    const Mock = struct {
        const vt = Command.VTable {
            .deinit = deinit_,
            .apply  = apply_,
            .revert = revert_,
        };
        count: i32 = 0,
        fn deinit_(_: *anyopaque) void { }
        fn apply_(ctx: *anyopaque) Command.Error!void {
            var self: *@This() = @ptrCast(@alignCast(ctx));
            self.count += 1;
        }
        fn revert_(ctx: *anyopaque) Command.Error!void {
            var self: *@This() = @ptrCast(@alignCast(ctx));
            self.count -= 1;
        }
    };
    var mock1 = Mock {};
    var mock2 = Mock {};

    var history = init(std.testing.allocator);
    defer history.deinit();

    try std.testing.expectEqual(false, history.isUndoAvailable());
    try std.testing.expectEqual(false, history.isRedoAvailable());
    try std.testing.expectError(error.NoCommand, history.undo());
    try std.testing.expectError(error.NoCommand, history.redo());

    try history.exec(Command { .ctx = &mock1, .vtable = &Mock.vt, });
    try std.testing.expectEqual(history.isUndoAvailable(), true);
    try std.testing.expectEqual(history.isRedoAvailable(), false);
    try std.testing.expectEqual(mock1.count, 1);

    try history.undo();
    try std.testing.expectEqual(history.isUndoAvailable(), false);
    try std.testing.expectEqual(history.isRedoAvailable(), true);
    try std.testing.expectEqual(mock1.count, 0);

    try history.redo();
    try std.testing.expectEqual(history.isUndoAvailable(), true);
    try std.testing.expectEqual(history.isRedoAvailable(), false);
    try std.testing.expectEqual(mock1.count, 1);

    try history.exec(Command { .ctx = &mock2, .vtable = &Mock.vt, });
    try std.testing.expectEqual(history.isUndoAvailable(), true);
    try std.testing.expectEqual(history.isRedoAvailable(), false);
    try std.testing.expectEqual(mock1.count, 1);
    try std.testing.expectEqual(mock2.count, 1);

    try history.undo();
    try std.testing.expectEqual(history.isUndoAvailable(), true);
    try std.testing.expectEqual(history.isRedoAvailable(), true);
    try std.testing.expectEqual(mock1.count, 1);
    try std.testing.expectEqual(mock2.count, 0);

    try history.redo();
    try std.testing.expectEqual(history.isUndoAvailable(), true);
    try std.testing.expectEqual(history.isRedoAvailable(), false);
    try std.testing.expectEqual(mock1.count, 1);
    try std.testing.expectEqual(mock2.count, 1);

    try history.undo();
    try std.testing.expectEqual(history.isUndoAvailable(), true);
    try std.testing.expectEqual(history.isRedoAvailable(), true);
    try std.testing.expectEqual(mock1.count, 1);
    try std.testing.expectEqual(mock2.count, 0);

    try history.undo();
    try std.testing.expectEqual(history.isUndoAvailable(), false);
    try std.testing.expectEqual(history.isRedoAvailable(), true);
    try std.testing.expectEqual(mock1.count, 0);
    try std.testing.expectEqual(mock2.count, 0);
}
